/*****************************************************************************
 * libvlc.c: main libvlc source
 * Includes the main() function for vlc. Parses command line, starts interface
 * and spawns threads.
 *****************************************************************************
 * Copyright (C) 1998-2001 VideoLAN
 * $Id: main.c,v 1.187 2002/04/24 00:36:24 sam Exp $
 *
 * Authors: Vincent Seguin <seguin@via.ecp.fr>
 *          Samuel Hocevar <sam@zoy.org>
 *          Gildas Bazin <gbazin@netcourrier.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 *****************************************************************************/

/*****************************************************************************
 * Pretend we are a builtin module
 *****************************************************************************/
#define MODULE_NAME main
#define __BUILTIN__

/*****************************************************************************
 * Preamble
 *****************************************************************************/
#include <errno.h>                                                 /* ENOMEM */
#include <stdio.h>                                              /* sprintf() */
#include <string.h>                                            /* strerror() */
#include <stdlib.h>                                                /* free() */
#include <signal.h>                               /* SIGHUP, SIGINT, SIGKILL */

#include <videolan/vlc.h>

#ifdef HAVE_GETOPT_LONG
#   ifdef HAVE_GETOPT_H
#       include <getopt.h>                                       /* getopt() */
#   endif
#else
#   include "GNUgetopt/getopt.h"
#endif

#ifndef WIN32
#   include <netinet/in.h>                            /* BSD: struct in_addr */
#endif

#ifdef HAVE_UNISTD_H
#   include <unistd.h>
#elif defined( _MSC_VER ) && defined( _WIN32 )
#   include <io.h>
#endif

#ifdef HAVE_LOCALE_H
#   include <locale.h>
#endif

#include "vlc_cpu.h"                                        /* CPU detection */

#include "netutils.h"                                 /* network_ChannelJoin */

#include "stream_control.h"
#include "input_ext-intf.h"

#include "intf_playlist.h"
#include "interface.h"

#include "audio_output.h"

#include "video.h"
#include "video_output.h"

#include "libvlc.h"

/*****************************************************************************
 * The evil global variables. We handle them with care, don't worry.
 *****************************************************************************/

/* This global lock is used for critical sections - don't abuse it! */
static vlc_mutex_t global_lock;
void *             p_global_data;

/* A list of all the currently allocated vlc objects */
static int i_vlc = 0;
static vlc_t **pp_vlc = NULL;

/*****************************************************************************
 * Local prototypes
 *****************************************************************************/
static int  GetFilenames  ( vlc_object_t *, int, char *[] );
static void Usage         ( vlc_object_t *, const char *psz_module_name );
static void ListModules   ( vlc_object_t * );
static void Version       ( void );
static void Build         ( void );

static void InitSignalHandler   ( void );
static void SimpleSignalHandler ( int i_signal );
static void FatalSignalHandler  ( int i_signal );

#ifdef WIN32
static void ShowConsole   ( void );
#endif

/*****************************************************************************
 * vlc_create: allocate a vlc_t structure, and initialize libvlc if needed.
 *****************************************************************************
 * This function allocates a vlc_t structure and returns NULL in case of
 * failure. Also, the thread system and the signal handlers are initialized.
 *****************************************************************************/
vlc_t * vlc_create( void )
{
    /* A root object that we use as a reference object */
    vlc_object_t root;

    /* Allocate the main structure */
    root.p_vlc = malloc( sizeof( vlc_t ) );

    if( root.p_vlc == NULL )
    {
        return NULL;
    }

    VLC_INIT_OBJECT( root.p_vlc, &root );
    root.p_vlc->p_global_lock = &global_lock;
    root.p_vlc->pp_global_data = &p_global_data;

    root.p_vlc->b_verbose = 0;
    root.p_vlc->b_quiet = 1; /* FIXME: delay message queue output! */

    /* Initialize the threads system */
    vlc_threads_init( root.p_vlc->p_this );

    /* Set signal handling policy for all threads */
    InitSignalHandler( );

    /* Store our newly allocated structure in the global list */
    vlc_mutex_lock( root.p_vlc->p_global_lock );
    pp_vlc = realloc( pp_vlc, (i_vlc+1) * sizeof( vlc_t * ) );
    pp_vlc[ i_vlc ] = root.p_vlc;
    i_vlc++;
    vlc_mutex_unlock( root.p_vlc->p_global_lock );

    return root.p_vlc;
}

/*****************************************************************************
 * vlc_init: initialize a vlc_t structure.
 *****************************************************************************
 * This function initializes a previously allocated vlc_t structure:
 *  - CPU detection
 *  - gettext initialization
 *  - message queue, module bank and playlist initialization
 *  - configuration and commandline parsing
 *****************************************************************************/
int vlc_init( vlc_t *p_vlc, int i_argc, char *ppsz_argv[] )
{
    char *psz_module;
    char *p_tmp;
    struct module_config_s *p_item;

    fprintf( stderr, COPYRIGHT_MESSAGE "\n" );

    /* Guess what CPU we have */
    p_vlc->i_cpu_capabilities = CPUCapabilities( p_vlc->p_this );

    /*
     * Support for gettext
     */
#if defined( ENABLE_NLS ) && defined ( HAVE_GETTEXT )
#   if defined( HAVE_LOCALE_H ) && defined( HAVE_LC_MESSAGES )
    if( !setlocale( LC_MESSAGES, "" ) )
    {
        fprintf( stderr, "warning: unsupported locale settings\n" );
    }

    setlocale( LC_CTYPE, "" );
#   endif

    if( !bindtextdomain( PACKAGE, LOCALEDIR ) )
    {
        fprintf( stderr, "warning: no domain %s in directory %s\n",
                 PACKAGE, LOCALEDIR );
    }

    textdomain( PACKAGE );
#endif

    /*
     * System specific initialization code
     */
    system_Init( p_vlc->p_this, &i_argc, ppsz_argv );

    /*
     * Initialize messages interface
     */
    msg_Create( p_vlc->p_this );

    /* Get the executable name (similar to the basename command) */
    if( i_argc > 0 )
    {
        p_vlc->psz_arg0 = p_tmp = ppsz_argv[ 0 ];
        while( *p_tmp )
        {
            if( *p_tmp == '/' ) p_vlc->psz_arg0 = ++p_tmp;
            else ++p_tmp;
        }
    }
    else
    {
        p_vlc->psz_arg0 = "vlc";
    }

    /*
     * Initialize the module bank and and load the configuration of the main
     * module. We need to do this at this stage to be able to display a short
     * help if required by the user. (short help == main module options)
     */
    module_InitBank( p_vlc->p_this );
    module_LoadMain( p_vlc->p_this );

    /* Hack: insert the help module here */
    help_module.psz_name = "help";
    help_module.psz_longname = _( "help module" );
    help_module.i_config_items =
                    sizeof(p_help_config) / sizeof(module_config_t) - 1;
    vlc_mutex_init( p_vlc->p_this, &help_module.config_lock );
    help_module.p_config = config_Duplicate( p_help_config );
    for( p_item = help_module.p_config;
         p_item->i_type != MODULE_CONFIG_HINT_END;
         p_item++ )
    {
        p_item->p_lock = &help_module.config_lock;
    }
    help_module.next = p_vlc->module_bank.first;
    p_vlc->module_bank.first = &help_module;
    /* End hack */

    if( config_LoadCmdLine( p_vlc->p_this, &i_argc, ppsz_argv, 1 ) )
    {
        msg_Destroy( p_vlc->p_this );
        return -1;
    }

    /* Check for short help option */
    if( config_GetInt( p_vlc->p_this, "help" ) )
    {
        fprintf( stderr, _("Usage: %s [options] [parameters] [file]...\n"),
                         p_vlc->psz_arg0 );

        Usage( p_vlc->p_this, "help" );
        Usage( p_vlc->p_this, "main" );
        return -1;
    }

    /* Check for version option */
    if( config_GetInt( p_vlc->p_this, "version" ) )
    {
        Version();
        return -1;
    }

    /* Check for build option */
    if( config_GetInt( p_vlc->p_this, "build" ) )
    {
        Build();
        return -1;
    }

    /* Hack: remove the help module here */
    p_vlc->module_bank.first = help_module.next;
    /* End hack */

    /*
     * Load the builtins and plugins into the module_bank.
     * We have to do it before config_Load*() because this also gets the
     * list of configuration options exported by each module and loads their
     * default values.
     */
    module_LoadBuiltins( p_vlc->p_this );
    module_LoadPlugins( p_vlc->p_this );
    msg_Info( p_vlc, "module bank initialized, found %i modules",
                     p_vlc->module_bank.i_count );

    /* Hack: insert the help module here */
    help_module.next = p_vlc->module_bank.first;
    p_vlc->module_bank.first = &help_module;
    /* End hack */

    /* Check for help on modules */
    if( (p_tmp = config_GetPsz( p_vlc->p_this, "module" )) )
    {
        Usage( p_vlc->p_this, p_tmp );
        free( p_tmp );
        return -1;
    }

    /* Check for long help option */
    if( config_GetInt( p_vlc->p_this, "longhelp" ) )
    {
        Usage( p_vlc->p_this, NULL );
        return -1;
    }

    /* Check for module list option */
    if( config_GetInt( p_vlc->p_this, "list" ) )
    {
        ListModules( p_vlc->p_this );
        return -1;
    }

    /* Hack: remove the help module here */
    p_vlc->module_bank.first = help_module.next;
    /* End hack */

    /*
     * Initialize mutexes
     */
    vlc_mutex_init( p_vlc->p_this, &p_vlc->config_lock );
    vlc_mutex_init( p_vlc->p_this, &p_vlc->structure_lock );

    /*
     * Override default configuration with config file settings
     */
    p_vlc->psz_homedir = config_GetHomeDir();
    config_LoadConfigFile( p_vlc->p_this, NULL );

    /*
     * Override configuration with command line settings
     */
    if( config_LoadCmdLine( p_vlc->p_this, &i_argc, ppsz_argv, 0 ) )
    {
#ifdef WIN32
        ShowConsole();
        /* Pause the console because it's destroyed when we exit */
        fprintf( stderr, "The command line options couldn't be loaded, check "
                 "that they are valid.\nPress the RETURN key to continue..." );
        getchar();
#endif
        msg_Destroy( p_vlc->p_this );
        return -1;
    }


    /*
     * System specific configuration
     */
    system_Configure( p_vlc->p_this );

    /* p_vlc inititalization. FIXME ? */
    p_vlc->i_desync = config_GetInt( p_vlc->p_this, "desync" ) * (mtime_t)1000;
    p_vlc->b_stats = config_GetInt( p_vlc->p_this, "stats" );
    p_vlc->b_verbose = config_GetInt( p_vlc->p_this, "verbose" );
    p_vlc->b_quiet = config_GetInt( p_vlc->p_this, "quiet" );
    p_vlc->b_audio = !config_GetInt( p_vlc->p_this, "noaudio" );
    p_vlc->b_stereo= !config_GetInt( p_vlc->p_this, "mono" );
    p_vlc->b_video = !config_GetInt( p_vlc->p_this, "novideo" );
    if( config_GetInt( p_vlc->p_this, "nommx" ) )
        p_vlc->i_cpu_capabilities &= ~CPU_CAPABILITY_MMX;
    if( config_GetInt( p_vlc->p_this, "no3dn" ) )
        p_vlc->i_cpu_capabilities &= ~CPU_CAPABILITY_3DNOW;
    if( config_GetInt( p_vlc->p_this, "nommxext" ) )
        p_vlc->i_cpu_capabilities &= ~CPU_CAPABILITY_MMXEXT;
    if( config_GetInt( p_vlc->p_this, "nosse" ) )
        p_vlc->i_cpu_capabilities &= ~CPU_CAPABILITY_SSE;
    if( config_GetInt( p_vlc->p_this, "noaltivec" ) )
        p_vlc->i_cpu_capabilities &= ~CPU_CAPABILITY_ALTIVEC;

    if( p_vlc->b_stats )
    {
        char p_capabilities[200];
        p_capabilities[0] = '\0';

#define PRINT_CAPABILITY( capability, string )                              \
        if( p_vlc->i_cpu_capabilities & capability )                       \
        {                                                                   \
            strncat( p_capabilities, string " ",                            \
                     sizeof(p_capabilities) - strlen(p_capabilities) );     \
            p_capabilities[sizeof(p_capabilities) - 1] = '\0';              \
        }

        PRINT_CAPABILITY( CPU_CAPABILITY_486, "486" );
        PRINT_CAPABILITY( CPU_CAPABILITY_586, "586" );
        PRINT_CAPABILITY( CPU_CAPABILITY_PPRO, "Pentium Pro" );
        PRINT_CAPABILITY( CPU_CAPABILITY_MMX, "MMX" );
        PRINT_CAPABILITY( CPU_CAPABILITY_3DNOW, "3DNow!" );
        PRINT_CAPABILITY( CPU_CAPABILITY_MMXEXT, "MMXEXT" );
        PRINT_CAPABILITY( CPU_CAPABILITY_SSE, "SSE" );
        PRINT_CAPABILITY( CPU_CAPABILITY_ALTIVEC, "AltiVec" );
        PRINT_CAPABILITY( CPU_CAPABILITY_FPU, "FPU" );
        msg_Info( p_vlc, "CPU has capabilities %s", p_capabilities );
    }

    /*
     * Initialize playlist and get commandline files
     */
    p_vlc->p_playlist = intf_PlaylistCreate( p_vlc->p_this );
    if( !p_vlc->p_playlist )
    {
        msg_Err( p_vlc, "playlist initialization failed" );
        msg_Destroy( p_vlc->p_this );
        return -1;
    }
    intf_PlaylistInit( p_vlc->p_playlist );

    /*
     * Get input filenames given as commandline arguments
     */
    GetFilenames( p_vlc->p_this, i_argc, ppsz_argv );

    /*
     * Initialize input, aout and vout banks
     */
    input_InitBank( p_vlc->p_this );
    aout_InitBank( p_vlc->p_this );
    vout_InitBank( p_vlc->p_this );

    /*
     * Choose the best memcpy module
     */
    psz_module = config_GetPsz( p_vlc->p_this, "memcpy" );
    p_vlc->p_memcpy_module =
        module_Need( p_vlc->p_this, MODULE_CAPABILITY_MEMCPY,
                     psz_module, NULL );
    if( psz_module ) free( psz_module );
    if( p_vlc->p_memcpy_module == NULL )
    {
        msg_Err( p_vlc, "no suitable memcpy module, using libc default" );
        p_vlc->pf_memcpy = memcpy;
    }
    else
    {
        p_vlc->pf_memcpy = p_vlc->p_memcpy_module->p_functions
                                  ->memcpy.functions.memcpy.pf_memcpy;
    }

    /*
     * Initialize shared resources and libraries
     */
    if( config_GetInt( p_vlc->p_this, "network-channel" )
         && network_ChannelCreate( p_vlc->p_this ) )
    {
        /* On error during Channels initialization, switch off channels */
        msg_Err( p_vlc,
                 "channels initialization failed, deactivating channels" );
        config_PutInt( p_vlc->p_this, "network-channel", 0 );
    }

    return 0;
}

/*****************************************************************************
 * vlc_run: start playing.
 *****************************************************************************
 * This function opens an interface plugin and runs it. If i_mode is set
 * to 0, vlc_run will return immediately and let the interface run in a
 * separate thread. If i_mode is set to 1, vlc_run will continue until user
 * requests to quit.
 *****************************************************************************/
int vlc_run( vlc_t *p_vlc, int i_mode )
{
    /* Try to run the interface */
    p_vlc->p_intf = intf_Create( p_vlc->p_this );

    if( p_vlc->p_intf == NULL )
    {
        msg_Err( p_vlc, "interface initialization failed" );
        return -1;
    }

    /* Launch the interface thread */
    p_vlc->p_intf->thread_id = intf_RunThread( p_vlc->p_intf );

    if( p_vlc->p_intf->thread_id == 0 )
    {
        intf_Destroy( p_vlc->p_intf );
        return -1;
    }

    /* If the user asked us to be modal, then let's be modal */
    if( i_mode == 1 )
    {
        while( !p_vlc->b_die )
        {
            msleep( INTF_IDLE_SLEEP );
        }
    }

    return 0;
}

/*****************************************************************************
 * vlc_stop: stop playing.
 *****************************************************************************
 * This function requests the interface thread to finish, waits for its
 * termination, and destroys its structure.
 *****************************************************************************/
int vlc_stop( vlc_t *p_vlc )
{
    /*
     * Ask the interface to stop
     */
    intf_StopThread( p_vlc->p_intf );

    /*
     * Finished, destroy the interface
     */
    intf_Destroy( p_vlc->p_intf );

    return 0;
}

/*****************************************************************************
 * vlc_end: uninitialize everything.
 *****************************************************************************
 * This function uninitializes every vlc component that was activated in
 * vlc_init: audio and video outputs, playlist, module bank and message queue.
 *****************************************************************************/
int vlc_end( vlc_t *p_vlc )
{
    /*
     * Go back into channel 0 which is the network
     */
    if( config_GetInt( p_vlc->p_this, "network-channel" ) && p_vlc->p_channel )
    {
        network_ChannelJoin( p_vlc->p_this, COMMON_CHANNEL );
    }

    /*
     * Free input, aout and vout banks
     */
    input_EndBank( p_vlc->p_this );
    vout_EndBank( p_vlc->p_this );
    aout_EndBank( p_vlc->p_this );

    /*
     * Free playlist
     */
    intf_PlaylistDestroy( p_vlc->p_playlist );

    /*
     * Free memcpy module if it was allocated
     */
    if( p_vlc->p_memcpy_module != NULL )
    {
        module_Unneed( p_vlc->p_memcpy_module );
    }

    /*
     * Free module bank
     */
    module_EndBank( p_vlc->p_this );

    /*
     * System specific cleaning code
     */
    system_End();

    /*
     * Destroy mutexes
     */
    vlc_mutex_destroy( &p_vlc->structure_lock );
    vlc_mutex_destroy( &p_vlc->config_lock );

    /*
     * Terminate messages interface and program
     */
    msg_Destroy( p_vlc->p_this );

    return 0;
}

/*****************************************************************************
 * vlc_destroy: free allocated resources.
 *****************************************************************************
 * This function frees the previously allocated vlc_t structure.
 *****************************************************************************/
int vlc_destroy( vlc_t *p_vlc )
{
    int i_index;

    /* Remove our structure from the global list */
    vlc_mutex_lock( p_vlc->p_global_lock );
    for( i_index = 0 ; i_index < i_vlc ; i_index++ )
    {
        if( pp_vlc[ i_index ] == p_vlc )
        {
            break;
        }
    }

    if( i_index == i_vlc )
    {
        fprintf( stderr, "error: trying to unregister %p which is not in "
                         "the list\n", p_vlc );
        vlc_mutex_unlock( p_vlc->p_global_lock );
        free( p_vlc );
        return -1;
    }

    for( i_index++ ; i_index < i_vlc ; i_index++ )
    {
        pp_vlc[ i_index - 1 ] = pp_vlc[ i_index ];
    }

    i_vlc--;
    pp_vlc = realloc( pp_vlc, i_vlc * sizeof( vlc_t * ) );
    vlc_mutex_unlock( p_vlc->p_global_lock );

    /* Stop thread system FIXME: last one out please shut the door! */
    //vlc_threads_end( );

    free( p_vlc );
    return 0;
}

/* following functions are local */

/*****************************************************************************
 * GetFilenames: parse command line options which are not flags
 *****************************************************************************
 * Parse command line for input files.
 *****************************************************************************/
static int GetFilenames( vlc_object_t *p_this, int i_argc, char *ppsz_argv[] )
{
    int i_opt;

    /* We assume that the remaining parameters are filenames */
    for( i_opt = optind; i_opt < i_argc; i_opt++ )
    {
        intf_PlaylistAdd( p_this->p_vlc->p_playlist, PLAYLIST_END,
                          ppsz_argv[ i_opt ] );
    }

    return 0;
}

/*****************************************************************************
 * Usage: print program usage
 *****************************************************************************
 * Print a short inline help. Message interface is initialized at this stage.
 *****************************************************************************/
static void Usage( vlc_object_t *p_this, const char *psz_module_name )
{
#define FORMAT_STRING "      --%s%s%s%s%s %s\n"
    /* option name -------------' | | | |  |
     * <bra ----------------------' | | |  |
     * option type or "" -----------' | |  |
     * ket> --------------------------' |  |
     * padding spaces ------------------'  |
     * comment ----------------------------'
     *
     * The purpose of having bra and ket is that we might i18n them as well.
     */
#define LINE_START 8
#define PADDING_SPACES 25
    module_t *p_module;
    module_config_t *p_item;
    char psz_spaces[PADDING_SPACES+LINE_START+1];
    char psz_format[sizeof(FORMAT_STRING)];

    memset( psz_spaces, ' ', PADDING_SPACES+LINE_START );
    psz_spaces[PADDING_SPACES+LINE_START] = '\0';

    strcpy( psz_format, FORMAT_STRING );

#ifdef WIN32
    ShowConsole();
#endif

    /* Enumerate the config of each module */
    for( p_module = p_this->p_vlc->module_bank.first ;
         p_module != NULL ;
         p_module = p_module->next )
    {
        if( psz_module_name && strcmp( psz_module_name, p_module->psz_name ) )
            continue;

        /* Ignore modules without config options */
        if( !p_module->i_config_items ) continue;

        /* Print module name */
        fprintf( stderr, _("%s module options:\n\n"), p_module->psz_name );

        for( p_item = p_module->p_config;
             p_item->i_type != MODULE_CONFIG_HINT_END;
             p_item++ )
        {
            char *psz_bra = NULL, *psz_type = NULL, *psz_ket = NULL;
            int i;

            if( p_item->i_short )
            {
                psz_format[2] = '-';
                psz_format[3] = p_item->i_short;
                psz_format[4] = ',';
            }
            else
            {
                psz_format[2] = ' ';
                psz_format[3] = ' ';
                psz_format[4] = ' ';
            }

            switch( p_item->i_type )
            {
            case MODULE_CONFIG_HINT_CATEGORY:
                fprintf( stderr, " %s\n", p_item->psz_text );
                break;

            case MODULE_CONFIG_ITEM_STRING:
            case MODULE_CONFIG_ITEM_FILE:
            case MODULE_CONFIG_ITEM_MODULE: /* We could also have "=<" here */
                psz_bra = " <"; psz_type = _("string"); psz_ket = ">";
                break;
            case MODULE_CONFIG_ITEM_INTEGER:
                psz_bra = " <"; psz_type = _("integer"); psz_ket = ">";
                break;
            case MODULE_CONFIG_ITEM_FLOAT:
                psz_bra = " <"; psz_type = _("float"); psz_ket = ">";
                break;
            case MODULE_CONFIG_ITEM_BOOL:
                psz_bra = ""; psz_type = ""; psz_ket = "";
                break;
            }

            if( psz_type )
            {
                i = PADDING_SPACES - strlen( p_item->psz_name )
                     - strlen( psz_bra ) - strlen( psz_type )
                     - strlen( psz_ket ) - 1;
                if( i < 0 )
                {
                    i = 0;
                    psz_spaces[i] = '\n';
                }
                else
                {
                    psz_spaces[i] = '\0';
                }

                fprintf( stderr, psz_format, p_item->psz_name, psz_bra,
                         psz_type, psz_ket, psz_spaces, p_item->psz_text );
                psz_spaces[i] = ' ';
            }
        }

        /* Yet another nasty hack.
         * Maybe we could use MODULE_CONFIG_ITEM_END to display tail messages
         * for each module?? */
        if( !strcmp( "main", p_module->psz_name ) )
        {
            fprintf( stderr, _("\nPlaylist items:"
                "\n  *.mpg, *.vob                   plain MPEG-1/2 files"
                "\n  [dvd:][device][@raw_device][@[title][,[chapter][,angle]]]"
                "\n                                 DVD device"
                "\n  [vcd:][device][@[title][,[chapter]]"
                "\n                                 VCD device"
                "\n  udpstream:[@[<bind address>][:<bind port>]]"
                "\n                                 UDP stream sent by VLS"
                "\n  vlc:loop                       loop execution of the "
                      "playlist"
                "\n  vlc:pause                      pause execution of "
                      "playlist items"
                "\n  vlc:quit                       quit VLC"
                "\n") );
        }

        fprintf( stderr, "\n" );

    }

#ifdef WIN32        /* Pause the console because it's destroyed when we exit */
        fprintf( stderr, _("\nPress the RETURN key to continue...\n") );
        getchar();
#endif
}

/*****************************************************************************
 * ListModules: list the available modules with their description
 *****************************************************************************
 * Print a list of all available modules (builtins and plugins) and a short
 * description for each one.
 *****************************************************************************/
static void ListModules( vlc_object_t *p_this )
{
    module_t *p_module;
    char psz_spaces[22];

    memset( psz_spaces, ' ', 22 );

#ifdef WIN32
    ShowConsole();
#endif

    /* Usage */
    fprintf( stderr, _("Usage: %s [options] [parameters] [file]...\n\n"),
                     p_this->p_vlc->psz_arg0 );

    fprintf( stderr, _("[module]              [description]\n") );

    /* Enumerate each module */
    for( p_module = p_this->p_vlc->module_bank.first ;
         p_module != NULL ;
         p_module = p_module->next )
    {
        int i;

        /* Nasty hack, but right now I'm too tired to think about a nice
         * solution */
        i = 22 - strlen( p_module->psz_name ) - 1;
        if( i < 0 ) i = 0;
        psz_spaces[i] = 0;

        fprintf( stderr, "  %s%s %s\n", p_module->psz_name, psz_spaces,
                  p_module->psz_longname );

        psz_spaces[i] = ' ';

    }

#ifdef WIN32        /* Pause the console because it's destroyed when we exit */
        fprintf( stderr, _("\nPress the RETURN key to continue...\n") );
        getchar();
#endif
}

/*****************************************************************************
 * Version: print complete program version
 *****************************************************************************
 * Print complete program version and build number.
 *****************************************************************************/
static void Version( void )
{
#ifdef WIN32
    ShowConsole();
#endif

    fprintf( stderr, VERSION_MESSAGE "\n" );
    fprintf( stderr,
      _("This program comes with NO WARRANTY, to the extent permitted by "
        "law.\nYou may redistribute it under the terms of the GNU General "
        "Public License;\nsee the file named COPYING for details.\n"
        "Written by the VideoLAN team at Ecole Centrale, Paris.\n") );

#ifdef WIN32        /* Pause the console because it's destroyed when we exit */
    fprintf( stderr, _("\nPress the RETURN key to continue...\n") );
    getchar();
#endif
}

/*****************************************************************************
 * Build: print information about the vlc build
 *****************************************************************************
 * Print the ./configure command line and other information.
 *****************************************************************************/
static void Build( void )
{
#ifdef WIN32
    ShowConsole();
#endif

    fprintf( stderr, "configured with %s\n", CONFIGURE_LINE );

#ifdef WIN32        /* Pause the console because it's destroyed when we exit */
    fprintf( stderr, _("\nPress the RETURN key to continue...\n") );
    getchar();
#endif
}

/*****************************************************************************
 * ShowConsole: On Win32, create an output console for debug messages
 *****************************************************************************
 * This function is useful only on Win32.
 *****************************************************************************/
#ifdef WIN32 /*  */
static void ShowConsole( void )
{
    AllocConsole();
    freopen( "CONOUT$", "w", stdout );
    freopen( "CONOUT$", "w", stderr );
    freopen( "CONIN$", "r", stdin );
    return;
}
#endif

/*****************************************************************************
 * InitSignalHandler: system signal handler initialization
 *****************************************************************************
 * Set the signal handlers. SIGTERM is not intercepted, because we need at
 * at least a method to kill the program when all other methods failed, and
 * when we don't want to use SIGKILL.
 *****************************************************************************/
static void InitSignalHandler( void )
{
#ifndef WIN32
    /* Termination signals */
    signal( SIGINT,  FatalSignalHandler );
    signal( SIGHUP,  FatalSignalHandler );
    signal( SIGQUIT, FatalSignalHandler );

    /* Other signals */
    signal( SIGALRM, SimpleSignalHandler );
    signal( SIGPIPE, SimpleSignalHandler );
#endif
}

/*****************************************************************************
 * SimpleSignalHandler: system signal handler
 *****************************************************************************
 * This function is called when a non fatal signal is received by the program.
 *****************************************************************************/
static void SimpleSignalHandler( int i_signal )
{
    int i_index;

    /* Acknowledge the signal received and warn all the p_vlc structures */
    vlc_mutex_lock( &global_lock );
    for( i_index = 0 ; i_index < i_vlc ; i_index++ )
    {
        msg_Warn( pp_vlc[ i_index ], "ignoring signal %d", i_signal );
    }
    vlc_mutex_unlock( &global_lock );
}

/*****************************************************************************
 * FatalSignalHandler: system signal handler
 *****************************************************************************
 * This function is called when a fatal signal is received by the program.
 * It tries to end the program in a clean way.
 *****************************************************************************/
static void FatalSignalHandler( int i_signal )
{
    int i_index;

    /* Once a signal has been trapped, the termination sequence will be
     * armed and following signals will be ignored to avoid sending messages
     * to an interface having been destroyed */
#ifndef WIN32
    signal( SIGINT,  SIG_IGN );
    signal( SIGHUP,  SIG_IGN );
    signal( SIGQUIT, SIG_IGN );
#endif

    /* Try to terminate everything - this is done by requesting the end of
     * all the p_vlc structures */
    vlc_mutex_lock( &global_lock );
    for( i_index = 0 ; i_index < i_vlc ; i_index++ )
    {
        /* Acknowledge the signal received */
        msg_Err( pp_vlc[ i_index ], "signal %d received, exiting", i_signal );
        pp_vlc[ i_index ]->b_die = 1;
    }
    vlc_mutex_unlock( &global_lock );
}

